import sys
import subprocess
from urllib.request import urlretrieve
from flumut.DbReader import get_db_file


def update():
    output = subprocess.check_output([sys.executable, '-m', 'pip', 'install', '-U', 'flumutdb'], stderr=subprocess.PIPE)


def update_db_file() -> None:
    url = 'https://github.com/izsvenezie-virology/FluMutDB/releases/latest/download/flumut_db.sqlite'
    db_path = get_db_file()
    _ = urlretrieve(url, db_path)



import unittest
import flumut.flumut as fm
import re


class TestNameParser(unittest.TestCase):

    def test_groups(self):
        pattern = re.compile(r'(.+)_(.+)')
        name = 'my_sample_name_PB1'
        result = fm.parse_name(name, pattern, False)
        self.assertEqual(result, ('my_sample_name', 'PB1'))

    def test_named_groups(self):
        pattern = re.compile(r'(?P<sample>.+)_(?P<segment>.+)')
        name = 'my_sample_name_PB1'
        result = fm.parse_name(name, pattern, False)
        self.assertEqual(result, ('my_sample_name', 'PB1'))        

    def test_named_groups_inverted(self):
        pattern = re.compile(r'(?P<segment>.+?)_(?P<sample>.+)')
        name = 'PB1_my_sample_name'
        result = fm.parse_name(name, pattern, False)
        self.assertEqual(result, ('my_sample_name', 'PB1'))        

    def test_unmatching_pattern(self):
        pattern = re.compile(r'unmatching_pattern')
        name = 'my_sample_name_PB1'
        with self.assertRaises(SystemExit) as ar:
            result = fm.parse_name(name, pattern, False)
    
    def test_unmatching_pattern_force(self):
        pattern = re.compile(r'unmatching_pattern')
        name = 'my_sample_name_PB1'
        result = fm.parse_name(name, pattern, True)
        self.assertEqual(result, (None, None))
    
    def test_no_groups(self):
        pattern = re.compile(r'.+')
        name = 'my_sample_name_PB1'
        with self.assertRaises(SystemExit) as ar:
            result = fm.parse_name(name, pattern, False)
    
    def test_no_groups_force(self):
        pattern = re.compile(r'.+')
        name = 'my_sample_name_PB1'
        result = fm.parse_name(name, pattern, True)
        self.assertEqual(result, (None, None))



import unittest
import flumut.flumut as fm


class TestTranslate(unittest.TestCase):

    def test_translate_all(self):
        sequence = 'TTTTTCTTATTGCTTCTCCTACTGATTATCATAATGGTTGTCGTAGTGTCTTCCTCATCGCCTCCCCCACCGACTACCACAACGGCTGCCGCAGCGTATTACTAATAGCATCACCAACAGAATAACAAAAAGGATGACGAAGAGTGTTGCTGATGGCGTCGCCGACGGAGTAGCAGAAGGGGTGGCGGAGGG---'
        result = fm.translate(sequence)
        self.assertEqual(result, ['F', 'F', 'L', 'L', 'L', 'L', 'L', 'L', 'I', 'I', 'I', 'M', 'V', 'V', 'V', 'V', 'S', 'S', 'S', 'S', 'P', 'P', 'P', 'P', 'T', 'T', 'T', 'T', 'A', 'A', 'A', 'A', 'Y', 'Y', '*', '*', 'H', 'H', 'Q', 'Q', 'N', 'N', 'K', 'K', 'D', 'D', 'E', 'E', 'C', 'C', '*', 'W', 'R', 'R', 'R', 'R', 'S', 'S', 'R', 'R', 'G', 'G', 'G', 'G', '-'])

    def test_translate_degen(self):
        sequence = 'RTTYTTSTTWTTKTTMTTBTTDTTHTTVTTNTT'
        result = fm.translate(sequence)
        self.assertEqual(result, ['IV', 'FL', 'LV', 'FI', 'FV', 'IL', 'FLV', 'FIV', 'FIL', 'ILV', '?'])

    def test_translate_del_inframe(self):
        sequence = 'ATG---CGT'
        result = fm.translate(sequence)
        self.assertEqual(result, ['M', '-', 'R'])

    def test_translate_del_frameshift(self):
        sequence = 'ATG-CGTTAC--ATG'
        result = fm.translate(sequence)
        self.assertEqual(result, ['M', 'R', 'Y', '-', 'M'])

    def test_translate_delayed_start(self):
        sequence = '---ATG---CGT'
        result = fm.translate(sequence)
        self.assertEqual(result, ['-', 'M', '-', 'R'])

    def test_translate_delayed_start_trunc(self):
        sequence = '--GATG---CGT'
        result = fm.translate(sequence)
        self.assertEqual(result, ['?', 'M', '-', 'R'])
        
    def test_translate_delayed_start_trunc_2(self):
        sequence = '----G---ATGC---C-GT'
        result = fm.translate(sequence)
        self.assertEqual(result, ['-', '?', '?', 'C', '-', 'R', '?'])

    pass



from collections import defaultdict
from typing import Any, Dict, List


class Mutation:
    def __init__(self, name: str, type: str, ref: str, alt: str, pos: int) -> None:
        self.name: str = name
        self.type: str = type
        self.ref: str = ref
        self.alt: str = alt
        self.pos: int = pos

        self.protein: str = self.name.split(':')[0]
        self.found: bool = False
        self.samples: Dict[str, List[str]] = {}


class Sample:
    def __init__(self, name: str) -> None:
        self.name: str = name
        self.segments: Dict[str, Dict[str, Any]] = defaultdict(dict)
        self.mutations: List[Mutation] = []
        self.markers: List[Dict[str, str]] = defaultdict(list)



import sys

class UnmatchNameException(Exception):
    def __init__(self, name, regex) -> None:
        self.name = name
        self.regex = regex
        self.message = f'Unable to parse "{name}" with regular expression "{regex}".'
        super().__init__(self.message)


class UnknownSegmentException(Exception):
    def __init__(self, name, regex, segment) -> None:
        self.name = name
        self.regex = regex
        self.message = f'Unrecognized segment "{segment}", found in "{name}" parsed with "{regex}".'
        super().__init__(self.message)


class UnknownNucleotideException(Exception):
    def __init__(self, codon) -> None:
        self.codon = codon
        self.message = f'Unexpected nucleotide in codon "{codon}".'
        super().__init__(self.message)

class MalformedFastaException(Exception):
    def __init__(self) -> None:
        self.message = 'Provided FASTA file does not start whit ">".'
        super().__init__(self.message)

class PermissionDeniedException(Exception):
    def __init__(self, file_name) -> None:
        self.file_name = file_name
        self.message = f'Permission denied while trying to write "{file_name}".'
        super().__init__(self.message)



import sqlite3
from importlib_resources import files
from typing import Any, Callable, Dict

_db_file: str = files('flumutdb').joinpath('flumut_db.sqlite')
_connection: sqlite3.Connection
_cursor: sqlite3.Cursor


def set_db_file(db_path: str) -> None:
    global _db_file
    _db_file = db_path


def get_db_file() -> str:
    return _db_file


def get_db_version():
    open_connection()
    major, minor, date = _cursor.execute('SELECT * FROM db_version').fetchone()
    close_connection()
    return major, minor, date


def open_connection() -> None:
    global _connection
    global _cursor
    _connection = sqlite3.connect(_db_file)
    _cursor = _connection.cursor()


def close_connection() -> None:
    global _connection
    global _cursor
    _connection.close()
    _connection = None
    _cursor = None


def execute_query(query: str, row_factory: Callable = None):
    _cursor.row_factory = row_factory
    return _cursor.execute(query)


def to_dict(cursor: sqlite3.Cursor, row: sqlite3.Row) -> Dict[str, Any]:
    result = {}
    for idx, col in enumerate(cursor.description):
        result[col[0]] = row[idx]
    return result



import csv
from importlib_resources import files
from typing import Dict, List, Tuple
from click.types import File
from openpyxl import Workbook, load_workbook
from openpyxl.worksheet.table import Table, TableStyleInfo
from openpyxl.utils.cell import get_column_letter

from flumut.DataClass import Mutation, Sample
from flumut.Exceptions import PermissionDeniedException


def mutations_dict(mutations: List[Mutation]) -> Tuple[List[str], List[Dict[str, str]]]:
    header = ['Sample']
    samples = {}
    for mutation in mutations:
        if mutation.found and mutation.name not in header:
            header.append(mutation.name)
        for sample in mutation.samples:
            if sample not in samples:
                samples[sample] = {'Sample': sample}
            samples[sample][mutation.name] = mutation.samples[sample]
    return header, samples.values()


def markers_dict(samples: List[Sample]) -> Tuple[List[str], List[Dict[str, str]]]:
    header = ['Sample', 'Marker', 'Mutations in your sample', 'Effect', 'Subtype', 'Literature']
    data = []
    for sample in samples:
        for marker in sample.markers:
            marker['Sample'] = sample.name
            data.append(marker)
    return header, data


def papers_dict(papers: List[Dict[str, str]]) -> Tuple[List[str], List[Dict[str, str]]]:
    header = ['Short name', 'Title', 'Authors', 'Year', 'Journal', 'Link', 'DOI']
    return header, papers


def write_csv(output_file: File, header: List[str], data: List[Dict[str, str]]) -> None:
    try:
        writer = csv.DictWriter(output_file, header, delimiter='\t', lineterminator='\n', extrasaction='ignore')
        writer.writeheader()
        writer.writerows(data)
    except PermissionError:
        raise PermissionDeniedException(output_file) from None


def get_workbook(vba: bool) -> Workbook:
    wb = load_workbook(files('flumutdata').joinpath('flumut_output.xlsm'), keep_vba=vba)
    wb['Markers per sample']._pivots[0].cache.refreshOnLoad = True
    return wb


def save_workbook(wb: Workbook, output_file: str) -> None:
    try:
        wb.save(output_file)
    except PermissionError:
        raise PermissionDeniedException(output_file) from None


def write_excel_sheet(wb: Workbook, sheet: str, header: List[str], data: List[Dict[str, str]]) -> Workbook:
    ws = wb[sheet]
    ws.append(header)
    for row, values in enumerate(data):
        for col, col_name in enumerate(header):
            ws.cell(row=row+2, column=col+1, value=values.get(col_name, ''))
    table = Table(displayName=f'{sheet}Table', ref=f'A1:{get_column_letter(len(header))}{len(data) + 1}')
    table.tableStyleInfo = TableStyleInfo(name="TableStyleMedium2", showFirstColumn=False, showLastColumn=False,
                                          showRowStripes=True, showColumnStripes=False)
    ws.add_table(table)
    return wb



__version__ = '0.6.1'
__author__ = 'Edoardo Giussani'
__contact__ = 'egiussani@izsvenezie.it'

from io import TextIOWrapper
import sys
from typing import Dict


def versions() -> Dict[str, str]:
    """Collect versions of FluMut and FluMutDB packages.

    Returns:
    versions : `Dict[str, str]`
            Versions for FluMut and FluMutDB packages.
    """
    from flumut.DbReader import get_db_version
    major, minor, date = get_db_version()
    return {
        'FluMut': __version__,
        'FluMutDB': f'{major}.{minor}, released on {date}'
    }


def update() -> None:
    """Updates FluMutDB package using Pip.
    """
    import flumut.Updater
    flumut.Updater.update()


def update_db_file() -> None:
    """Updates FluMutDB database without updating 
    """
    import flumut.Updater
    flumut.Updater.update_db_file()


def analyze(name_regex: str, fasta_file: TextIOWrapper, db_file: str,
            markers_output: TextIOWrapper, mutations_output: TextIOWrapper, literature_output: TextIOWrapper, excel_output: str,
            relaxed: bool = False, skip_unmatch_names: bool = False, skip_unknown_segments: bool = False, 
            debug: bool = False, verbose: bool = False) -> None:
    """Runs the FluMut analysis.

    Args:
        name_regex: `str`
            The regex used to parse FASTA header
        fasta_file: `TextIOWrapper`
            The opened FASTA file to analyze
        db_file: `str|None`
            Path of the db file. If `None` the file from FluMutDB is used
        markers_output: `TextIOWrapper|None`
            The opened file where write Markers output
        mutations_output: `TextIOWrapper|None`
            The opened file where write Mutations output
        literature_output: `TextIOWrapper|None`
            The opened file where write Literature output
        excel_output: `str`
            The complete path to the Excel output file
        relaxed: `bool`
            When `True`, all markers where at least one mutation is detected
        skip_unmatch_names: `bool`
            When `True`, unmatching names do not raise exceptions
        skip_unknown_segments: `bool`
            When `True`, unknown segments do not raise exceptions
    """

    if not debug:
        sys.tracebacklimit = 0

    if name_regex is None:
        name_regex = r'(?P<sample>.+)_(?P<segment>.+)'
    if db_file is not None:
        from flumut.DbReader import set_db_file
        set_db_file(db_file)

    import flumut.flumut
    flumut.flumut.analyze(name_regex=name_regex,
                          fasta_file=fasta_file,
                          markers_output=markers_output,
                          mutations_output=mutations_output,
                          literature_output=literature_output,
                          excel_output=excel_output,
                          relaxed=relaxed,
                          skip_unmatch_names=skip_unmatch_names,
                          skip_unknown_segments=skip_unknown_segments,
                          verbose=verbose)



import click
from click import File

import flumut
from flumut import __version__, __author__, __contact__


def update(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    old_version = flumut.versions()['FluMutDB']
    flumut.update()
    new_version = flumut.versions()['FluMutDB']
    if old_version == new_version:
        print(f'Already using latest FluMutDB version ({new_version})')
    else:
        print(f'Updated FluMutDB to version {new_version}')
    ctx.exit()


def versions(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    for package, version in flumut.versions().items():
        print(f'{package}: {version}')
    ctx.exit()


@click.command()
@click.help_option('-h', '--help')
@click.version_option(__version__, '-v', '--version', message=f'%(prog)s, version %(version)s, by {__author__} ({__contact__})')
@click.option('-V', '--all-versions', is_flag=True, callback=versions, expose_value=False, is_eager=True, help='Prints all versions and exit.')
@click.option('--update', is_flag=True, callback=update, expose_value=False, is_eager=True, help='Update the database to the latest version and exit.')
@click.option('--skip-unmatch-names', is_flag=True, default=False, help='Skip sequences with name that does not match the regular expression pattern.')
@click.option('--skip-unknown-segments', is_flag=True, default=False, help='Skip sequences with segment not present in the database.')
@click.option('-r', '--relaxed', is_flag=True, help='Report markers of which at least one mutation is found.')
@click.option('-n', '--name-regex', type=str, default=r'(?P<sample>.+)_(?P<segment>.+)', show_default=True, help='Set regular expression to parse sequence name.')
@click.option('-D', '--db-file', type=str, default=None, help='Set source database.')
@click.option('-m', '--markers-output', type=File('w', 'utf-8'), default=None, help='TSV markers output file.')
@click.option('-M', '--mutations-output', type=File('w', 'utf-8'), default=None, help='TSV mutations output file.')
@click.option('-l', '--literature-output', type=File('w', 'utf-8'), default=None, help='TSV literature output file.')
@click.option('-x', '--excel-output', type=str, default=None, help='Excel complete report.')
@click.option('--debug', is_flag=True, hidden=True, help='Output errors with traceback')
@click.option('--verbose', is_flag=True, hidden=True)
@click.argument('fasta-file', type=File('r'))
def cli(name_regex: str, fasta_file: File, db_file: str,
        markers_output: File, mutations_output: File, literature_output: File, excel_output: str,
        relaxed: bool, skip_unmatch_names: bool, skip_unknown_segments: bool, debug: bool, verbose: bool) -> None:
    flumut.analyze(**locals())


if __name__ == '__main__':
    cli()



#! /usr/bin/env python

import re
import sys
import operator
import itertools

from io import TextIOWrapper
from typing import Dict, Generator, List, Optional, Tuple

from collections import defaultdict
from Bio.Align import PairwiseAligner

from flumut.DbReader import close_connection, execute_query, open_connection, to_dict
from flumut import OutputFormatter
from flumut.DataClass import Mutation, Sample
from flumut.Exceptions import UnmatchNameException, UnknownSegmentException, UnknownNucleotideException, MalformedFastaException

PRINT_ALIGNMENT = False


def analyze(name_regex: str, fasta_file: TextIOWrapper,
            markers_output: TextIOWrapper, mutations_output: TextIOWrapper, literature_output: TextIOWrapper, excel_output: str,
            relaxed: bool, skip_unmatch_names: bool, skip_unknown_segments: bool, verbose: bool) -> None:
    '''
    Find markers of zoonotic interest in H5N1 avian influenza viruses.
    '''

    # Initialization
    samples: Dict[str, Sample] = {}
    pattern = re.compile(name_regex)
    counter = 0

    open_connection()
    segments = load_segments()
    mutations = load_mutations()
    annotations = load_annotations()
    close_connection()

    # Per sequence analysis
    for name, seq in read_fasta(fasta_file):
        if verbose:
            print(f'Processed sequences : {counter}', file=sys.stderr)
        counter += 1

        sample,  segment = parse_name(name, pattern, skip_unmatch_names)
        if sample is None or segment is None:
            continue
        if segment not in segments:
            ex = UnknownSegmentException(name, pattern.pattern, segment)
            if not skip_unknown_segments:
                raise ex
            print(ex.message, file=sys.stderr)
            continue

        if sample not in samples:
            samples[sample] = Sample(sample)

        reference_name, reference_sequence = select_reference(segments[segment], seq)
        ref_align, sample_align = pairwise_alignment(reference_sequence, seq)

        for protein, cds in annotations[reference_name].items():
            ref_cds, sample_cds = get_cds(ref_align, sample_align, cds)
            ref_aa = ''.join(translate(ref_cds))
            sample_aa = translate(sample_cds)

            samples[sample].mutations += find_mutations(
                ref_aa, sample_aa, sample, mutations[protein])

    open_connection()
    for sample in samples.values():
        sample.markers = match_markers(sample.mutations, relaxed)
    papers = load_papers()
    close_connection()
    found_mutations = list(itertools.chain.from_iterable(mutations.values()))
    found_mutations.sort(key=operator.attrgetter('protein', 'pos', 'alt'))

    if verbose:
        print('Writing outputs', file=sys.stderr)

    # Outputs
    if mutations_output:
        header, data = OutputFormatter.mutations_dict(found_mutations)
        OutputFormatter.write_csv(mutations_output, header, data)

    if markers_output:
        header, data = OutputFormatter.markers_dict(samples.values())
        OutputFormatter.write_csv(markers_output, header, data)

    if literature_output:
        header, data = OutputFormatter.papers_dict(papers)
        OutputFormatter.write_csv(literature_output, header, data)

    if excel_output:
        wb = OutputFormatter.get_workbook(excel_output.endswith('.xlsm'))
        header, data = OutputFormatter.mutations_dict(found_mutations)
        wb = OutputFormatter.write_excel_sheet(wb, 'Mutations', header, data)
        header, data = OutputFormatter.markers_dict(samples.values())
        wb = OutputFormatter.write_excel_sheet(wb, 'Markers', header, data)
        header, data = OutputFormatter.papers_dict(papers)
        wb = OutputFormatter.write_excel_sheet(wb, 'Literature', header, data)
        wb = OutputFormatter.save_workbook(wb, excel_output)


def load_mutations() -> Dict[str, List[Mutation]]:
    mutations = defaultdict(list)
    res = execute_query(""" SELECT reference_name, protein_name, name, type, ref_seq, alt_seq, position
                            FROM mutation_mappings
                            JOIN mutations ON mutation_mappings.mutation_name = mutations.name""")
    for mut in res:
        mutations[mut[1]].append(Mutation(*mut[2:]))
    return mutations


def load_segments() -> Dict[str, Dict[str, str]]:
    res = execute_query("SELECT segment_name, name, sequence FROM 'references'")
    segments = defaultdict(dict)
    for segment, name, sequence in res:
        segments[segment][name] = sequence
    return segments


def load_annotations() -> Dict[str, Dict[str, List[Tuple[int, int]]]]:
    res = execute_query("SELECT reference_name, protein_name, start, end FROM 'annotations'")
    ann = defaultdict(lambda: defaultdict(list))
    for ref, prot, start, end in res:
        ann[ref][prot].append((start, end))
    return ann


def load_papers() -> List[Dict[str, str]]:
    return execute_query("""SELECT  id AS 'Short name',
                                    title AS 'Title',
                                    authors AS 'Authors',
                                    year AS 'Year',
                                    journal AS 'Journal',
                                    web_address AS 'Link',
                                    doi AS 'DOI'
                            FROM papers""", to_dict).fetchall()


def match_markers(muts: List[Mutation], relaxed: bool) -> List[Dict[str, str]]:
    muts_str = ','.join([f"'{mut.name}'" for mut in muts])
    res = execute_query(f"""
    WITH markers_tbl AS (SELECT marker_id,
                                group_concat(mutation_name) AS found_mutations,
                                count(mutation_name) AS found_mutations_count
                            FROM markers_mutations
                            WHERE mutation_name IN ({muts_str})
                            GROUP BY markers_mutations.marker_id)

    SELECT  markers_summary.all_mutations AS 'Marker',
            markers_tbl.found_mutations AS 'Mutations in your sample',
            markers_effects.effect_name AS 'Effect', 
            group_concat(markers_effects.paper_id, '; ') AS 'Literature', 
            markers_effects.subtype AS 'Subtype'
    FROM markers_effects
    JOIN markers_tbl ON markers_tbl.marker_id = markers_effects.marker_id
    JOIN markers_summary ON markers_summary.marker_id = markers_effects.marker_id
    WHERE markers_effects.marker_id IN (
        SELECT markers_tbl.marker_id 
        FROM markers_tbl) { 'AND markers_summary.all_mutations_count = markers_tbl.found_mutations_count' if not relaxed else '' }
    GROUP BY markers_effects.marker_id, markers_effects.effect_name, markers_effects.subtype
    """, to_dict)
    return res.fetchall()


def select_reference(references: Dict[str, str], ref_seq: str) -> Tuple[str, str]:
    if len(references) > 1:
        NotImplementedError('Selection for reference from segments with more than one is not yet implemented')
    (name, sequence), = references.items()
    return name, sequence


def find_mutations(ref_aa: str, sample_aa: List[str], sample_name: str, mutations: List[Mutation]):
    found_mutations = []
    for mutation in mutations:
        pos = adjust_position(ref_aa, mutation.pos)
        mutation.samples[sample_name] = sample_aa[pos]
        if mutation.alt in sample_aa[pos]:
            mutation.found = True
            found_mutations.append(mutation)
    return found_mutations


def pairwise_alignment(ref_seq: str, sample_seq: str) -> Tuple[str, str]:
    '''Align sequence against a reference'''
    aligner = PairwiseAligner()
    aligner.open_gap_score = -10
    aligner.extend_gap_score = -1
    alignment = aligner.align(ref_seq, sample_seq)[0]
    if PRINT_ALIGNMENT:
        print(alignment, file=sys.stderr)
    return alignment[0], alignment[1]


def read_fasta(fasta_file: TextIOWrapper) -> Generator[str, None, None]:
    '''Create a Fasta reading a file in Fasta format'''
    name = None
    for line in fasta_file:
        if line.startswith('>'):
            if name is not None:
                yield name, ''.join(seq).upper()
            name = line[1:].strip()
            seq = []
        else:
            try:
                seq.append(line.strip())
            except UnboundLocalError:
                raise MalformedFastaException() from None
    if name is not None:
        yield name, ''.join(seq).upper()


def parse_name(name: str, pattern: re.Pattern, force: bool) -> Tuple[str, str]:
    '''Get sample and segment information by sequence name'''
    match = pattern.match(name)
    try:
        sample = match.groupdict().get('sample', match.group(1))
        seg = match.groupdict().get('segment', match.group(2))
    except (IndexError, AttributeError):
        ex = UnmatchNameException(name, pattern.pattern)
        if not force:
            raise ex from None
        print(ex.message, file=sys.stderr)
        return None, None
    else:
        return sample, seg


def translate(seq: str) -> List[str]:
    '''Translate nucleotidic sequence in AA sequence'''
    nucls = list(seq)
    aas = []
    is_first = True

    for i in range(0, len(nucls), 3):
        codon = get_codon(nucls, i, is_first)
        aa = translate_codon(codon)
        if aa not in ('-', '?'):
            is_first = False
        aas.append(aa)
    return aas


def get_codon(seq: List[str], start: int, is_first: bool) -> List[str]:
    '''Exctract the codon'''
    codon = seq[start:start + 3]
    if codon == ['-', '-', '-']:  # If the codon is a deletion
        return codon
    # If the codon starts from mid codon (to avoid frameshifts in truncated sequences):
    if is_first and codon[0] == '-':
        return codon
    codon = [n for n in codon if n != '-']
    while len(codon) < 3:
        next_nucl = find_next_nucl(seq, start)
        if not next_nucl:
            break
        codon.append(seq[next_nucl])
        seq[next_nucl] = '-'
    return codon


def translate_codon(codon: List[str]) -> str:
    '''Translate a codon into a set of AAs, containing all possible combinations in case of degenerations'''
    if 'N' in codon:
        return '?'
    try:
        undegenerated_codon = [degeneration_dict[nucl] for nucl in codon]
    except KeyError:
        raise UnknownNucleotideException(''.join(codon))
    codons = list(itertools.product(*undegenerated_codon))
    aas = [translation_dict.get(''.join(c), '?') for c in codons]
    return ''.join(sorted(set(aas)))


def find_next_nucl(seq: List[str], start: int) -> Optional[int]:
    '''Returns the position of the next non deleted nucleotide'''
    for i in range(start + 3, len(seq)):
        if not seq[i] == '-':
            return i
    return None


def get_cds(ref_seq: str, sample_seq: str, cds: List[Tuple[int, int]]) -> Tuple[str, str]:
    '''Cut and assemble the nucleotide sequences based on positions given by the cds'''
    cds.sort(key=lambda x: x[0])
    ref_nucl = ''
    sample_nucl = ''

    for rng in cds:
        start = adjust_position(ref_seq, rng[0])
        end = adjust_position(ref_seq, rng[1]) + 1
        ref_nucl += ref_seq[start:end]
        sample_nucl += sample_seq[start:end]
    return ref_nucl, sample_nucl


def adjust_position(ref_seq: str, pos: int) -> int:
    '''Adjust 1-based position to 0-based, considering reference sequence gaps'''
    pos -= 1    # conversion 1-based to 0-based numeration
    dashes = 0
    adj_pos = pos
    while ref_seq.count('-', 0, adj_pos + 1) != dashes:
        dashes = ref_seq.count('-', 0, adj_pos + 1)
        adj_pos = pos + dashes
    return adj_pos


translation_dict = {
    'TTT': 'F', 'TTC': 'F', 'TTA': 'L', 'TTG': 'L', 'CTT': 'L', 'CTC': 'L', 'CTA': 'L', 'CTG': 'L',
    'ATT': 'I', 'ATC': 'I', 'ATA': 'I', 'ATG': 'M', 'GTT': 'V', 'GTC': 'V', 'GTA': 'V', 'GTG': 'V',
    'TCT': 'S', 'TCC': 'S', 'TCA': 'S', 'TCG': 'S', 'CCT': 'P', 'CCC': 'P', 'CCA': 'P', 'CCG': 'P',
    'ACT': 'T', 'ACC': 'T', 'ACA': 'T', 'ACG': 'T', 'GCT': 'A', 'GCC': 'A', 'GCA': 'A', 'GCG': 'A',
    'TAT': 'Y', 'TAC': 'Y', 'TAA': '*', 'TAG': '*', 'CAT': 'H', 'CAC': 'H', 'CAA': 'Q', 'CAG': 'Q',
    'AAT': 'N', 'AAC': 'N', 'AAA': 'K', 'AAG': 'K', 'GAT': 'D', 'GAC': 'D', 'GAA': 'E', 'GAG': 'E',
    'TGT': 'C', 'TGC': 'C', 'TGA': '*', 'TGG': 'W', 'CGT': 'R', 'CGC': 'R', 'CGA': 'R', 'CGG': 'R',
    'AGT': 'S', 'AGC': 'S', 'AGA': 'R', 'AGG': 'R', 'GGT': 'G', 'GGC': 'G', 'GGA': 'G', 'GGG': 'G',
    '---': '-'
}


degeneration_dict = {
    'A': ['A'], 'C': ['C'], 'G': ['G'], 'T': ['T'], 'U': ['T'], '-': ['-'],
    'R': ['A', 'G'], 'Y': ['C', 'T'], 'S': ['G', 'C'], 'W': ['A', 'T'],
    'K': ['G', 'T'], 'M': ['A', 'C'], 'B': ['C', 'G', 'T'], 'D': ['A', 'G', 'T'],
    'H': ['A', 'C', 'T'], 'V': ['A', 'C', 'G'], 'N': ['A', 'C', 'G', 'T']
}



import unittest
import flumut.flumut as fm


class TestPosition(unittest.TestCase):

    def test_adjust_normal(self):
        sequence = 'ACTG'
        result = fm.adjust_position(sequence, 1)
        self.assertEqual(result, 0)
        
    def test_adjust_end(self):
        sequence = 'ACTG'
        result = fm.adjust_position(sequence, 4)
        self.assertEqual(result, 3)
        
    def test_adjust_gap(self):
        sequence = '-CTG'
        result = fm.adjust_position(sequence, 1)
        self.assertEqual(result, 1)
        
    def test_adjust_gap_long(self):
        sequence = '---AT--GC'
        result = fm.adjust_position(sequence, 1)
        self.assertEqual(result, 3)
        
    def test_adjust_gap_multiple(self):
        sequence = '---AT-----GC'
        result = fm.adjust_position(sequence, 4)
        self.assertEqual(result, 11)
